#!/usr/bin/python
# -*- coding: utf-8 -*-

# Description: This is the implementation of a mapwoc-Redirector.
#
# Copyright:
# Copyright (C) 2011 Bundesamt fur Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

from optparse import OptionParser
import socket
import threading
import time
from copy import deepcopy
import logging
import ConfigParser

from sqlalchemy import *

from mapwoc.lib.url import URL, URLList, validate_URL
from mapwoc.lib.toolutil import print_version_exit
from mapwoc.lib.redirectorvisit import RedirectorVisit
from mapwocgui.model.meta import Session
import mapwoc.lib.model.url

from redirectorserver import RedirectorServer
from redirectorctrlserver import RedirectorCtrlServer

logger = logging.getLogger('redirector')

class RedirectorState:
    """State 'enum'."""

    IDLE = 0 # Not serving URLs
    RUNNING = 200 # Serving URLs
    DONE = 201 # If looping, serving URLs, served each already at least once.
    LOADING = 300 # Loading URL List
    ERROR = 400 # Error occurred


class RedirectorMode:
    """Operation-mode 'enum'."""

    COMPREHENSIVE = 100 # One list per hcsid.
    SAMPLE = 200        # All hcsids draw from the same list.


def mapping_from_file(urllist_file):
    """Returns a dict with keys being hcsid and values a list of urls
       that this hcs shall visit.
       If in sample mode, uses an empty string as key.
    """

    url_mapping = {}
    curr_section = ''
    for line in open(urllist_file, 'r').readlines():
        if line.startswith("["):
            curr_section = line[1:-2]
        else:
            urls = url_mapping.get(curr_section, [])
            if not validate_URL(line.strip()):
                logger.warning("Ignoring invalid URL: %s from list %s " % (
                    line.strip(), urllist_file))
                continue
            urls += [line.strip()]
            url_mapping[curr_section] = urls
    return url_mapping


def mapping_from_db(parameters):
    """Return a mapping from database."""

    parameters = parameters.split(",")
    database = parameters.pop()

    try:
        logger.debug("Opening database %s" % database)
        engine = create_engine('sqlite:///%s' % database)
        c = engine.connect()
        Session.configure(bind=engine)
    except:
        logger.warning("Unable to open database %s " % database)
        return None

    urllists = []
    for listname in parameters:
        # Ignore empty listnames
        if not listname: continue
        logger.debug("looking for listname: %s" % listname)
        urllist = Session.query(URLList).filter(URLList.name == listname).first()
        if urllist is None:
            logger.warning("Could not find an urllist with name '%s' in DB." % listname)
        else:
            logger.info("Loaded urllist '%s' with %d URLs. " %
                                (urllist.name, len(urllist.urls)))
            urllists.append(urllist)

    if not urllists:
        logger.warning("Could not load any urllists")
        return None

    # TODO we default to sample mode here.

    mapping = {}
    # TODO use listdict or better python-fu here.
    mapping[''] = []

    for urllist in urllists:
        for url in urllist.urls:
            if not validate_URL(url.url):
                logger.warning("Invalid URL %s in database list: %s" % (
                    url.url, name))
                continue
            mapping[''].append(url.url)

    return mapping


# Using Threads and the servers as done here is a bad design decision, but ok for prototype.
# Use multiprocessing in the future, and/or a twisted-like approach.

class Redirector:
    """Redirector, serves redirects and listens on controll-socket."""

    VERSION = "0.0.1"

    def __init__(self, redirector_port=8338, ctrl_port=8228) :
        """Prepares the servers to listen for http and control requests.

           redirector_port Port on which Redirector listens for http requests.
           ctrl_port Port on which ControlServer listens for commands.
        """

        self.mapping = {}
        self.serve_history = []
        self.redirect_port = redirector_port
        self.ctrl_port = ctrl_port
        self.redirectorsrv = RedirectorServer(self,
                                     socket.gethostname(), self.redirect_port)
        self.ctrlsrv = RedirectorCtrlServer(self,
                                            ('localhost', self.ctrl_port))
        self.mode = RedirectorMode.SAMPLE
        self.state = RedirectorState.IDLE
        self.do_loop = True
        self.mapping_to_loop = deepcopy(self.mapping)
        self.__num_urls_per_request = 1

    def is_fast_mode(self):
        """Returns True if one url should be given per request."""
        return self.__num_urls_per_request != 1

    def set_num_urls(self, numurls):
        self.__num_urls_per_request = numurls

    def get_num_urls(self):
        return self.__num_urls_per_request

    def start(self, mapping=None) :
        """Starts the server with given mapping (blocks)."""

        self.mapping = mapping
        self.mapping_to_loop = deepcopy(self.mapping)
        self.__update_mode()

        try:
            redirectorsrv_thread = threading.Thread(
                                       target=self.redirectorsrv.serve_forever)
            redirectorsrv_thread.setDaemon(True)
            redirectorsrv_thread.start()
            logger.info("Redirector listening at port %d." % self.redirect_port)
            redirectorctrlsrv_thread = threading.Thread(
                                        target=self.ctrlsrv.serve_forever)
            redirectorctrlsrv_thread.setDaemon(True)
            redirectorctrlsrv_thread.start()
            logger.info("Redirector-Ctrl listening at port %d." % self.ctrl_port)
        except KeyboardInterrupt:
            pass

        redirectorsrv_thread.join()
        self.ctrlsrv.shutdown()
        redirectorctrlsrv_thread.join()


    def get_next_url(self, hcsid, peer):
        """Get the next URL for a given hcsid, None if none left."""
      
        # Allow queries that are identifiable by hcsid also in sample mode.

        if self.mode == RedirectorMode.SAMPLE:
            query_hcsid = ''
        else:
            query_hcsid = hcsid

        url = None
        try:
            url = self.mapping[query_hcsid].pop()
        except KeyError:
            self.state = RedirectorState.DONE
        except IndexError:
            if (self.mode == RedirectorMode.SAMPLE) and (self.do_loop == True):
                self.restart_with_mapping(self.mapping_to_loop)
                # Agreed that this can be done much nicer.
                url = self.mapping[query_hcsid].pop()
                self.state = RedirectorState.DONE

        # TODO That means an IDLE Redirector will become DONE after
        #      a request. Fix that.
        if self.mode == RedirectorMode.SAMPLE:
            remains = self.mapping.get('')
            if remains is None or len(remains) == 0:
                self.state = RedirectorState.DONE

        self.log_visit(hcsid, url, peer)
        if url is None:
            return None
        return url.strip()


    def peek_next_url(self, hcsid, n=1):
        """Look at the next URL for a given hcsid, empty list in case of errors."""

        try:
            if len(self.mapping) > 0:
                upcoming_urls = self.mapping[hcsid]
                if upcoming_urls is None:
                    return []
                last = upcoming_urls[-n:] 
                last.reverse()
                return last
            else:
                return []
        except KeyError:
            return []


    def stop(self):
        """Attempt to stop redirector server."""
        
        self.redirectorsrv.shutdown()


    def log_visit(self, hcsid, url, peer):
        """Adds an entry to the serve_history."""

        # Only log URLs.
        if url is not None:
            self.serve_history += [RedirectorVisit(hcsid, url, peer)]

    def go_idle(self):
        """Loads empty mapping and set own state to idle."""
        self.restart_with_mapping({'':[]})
        self.state = RedirectorState.IDLE

    def str_state(self):
        """Return a simple textual representation of current state."""

        if self.state == RedirectorState.IDLE:
            return 'idle'
        elif self.state == RedirectorState.RUNNING:
            return 'running'
        elif self.state == RedirectorState.LOADING:
            return 'loading'
        elif self.state == RedirectorState.DONE:
            return 'done'
        else:
            return 'error'


    def load_url_file(self, filepath):
        """Load an url list from file, returning True if
           successfull (False otherwise).
        """

        self.state = RedirectorState.LOADING
        try:
            mapping = mapping_from_file(filepath)
            self.restart_with_mapping(mapping)
            self.state = RedirectorState.RUNNING
            return True
        except IOError:
            self.state = RedirectorState.ERROR
            return False


    def restart_with_mapping(self, mapping):
        """Sets a new mapping and state to RUNNING."""
        
        self.state = RedirectorState.LOADING
        self.mapping = mapping
        self.__update_mode()
        self.mapping_to_loop = deepcopy(mapping)
        self.state = RedirectorState.RUNNING
        # Reset the history.
        self.serve_history = []


    def __update_mode(self):
        """Sets mode variable according to the mapping."""

        if len(self.mapping.keys()) == 1 and '' in self.mapping.keys():
            self.mode = RedirectorMode.SAMPLE
        else:
            self.mode = RedirectorMode.COMPREHENSIVE

    def load_url_db(self, parameters):
        """
        Expects a colon seperated list of parameters that
        contain the names of the lists that should be loaded.
        The last element needs to be the database from which
        those urls should be loaded.

            Return True if successfull, false otherwise.
        """

        ul = mapping_from_db(parameters)
        if ul is not None:
            self.restart_with_mapping(ul)
        else:
            return False
        return True


def setup_logging(quiet, verbose):
    """Setup logging."""

    mainlogger = logging.getLogger('redirector')
    mainlogger.addHandler(logging.FileHandler('redirector.log'))
    mainlogger.addHandler(logging.StreamHandler())
    ctrllogger = logging.getLogger('redirectorctrl')
    ctrllogger.addHandler(logging.FileHandler('redirector.log'))
    ctrllogger.addHandler(logging.StreamHandler())
    serverlogger = logging.getLogger('redirectorsrv')
    serverlogger.addHandler(logging.FileHandler('redirector.log'))
    serverlogger.addHandler(logging.StreamHandler())

    min_level = logging.INFO

    if verbose:
        min_level = logging.DEBUG
    elif quiet:
        min_level = logging.WARNING

    mainlogger.setLevel(min_level)
    ctrllogger.setLevel(min_level)
    serverlogger.setLevel(min_level)

def main():
    """Starts a redirector, serving the URLs found in a file.
       The redirector listens on a second interface for control commands.
    """

    parser = OptionParser()
    parser.add_option("", "--urllist", dest="urllistfile", help="Path to file containing URLs.")
    parser.add_option("", "--db", dest="db", help="Path to database (ignored).")
    parser.add_option("", "--config", dest="config",
                      help="Path to file containing config.",
                      default="/etc/mapwoc/redirector/redirector.config")
    parser.add_option("-v", "--verbose", dest="verbose", action="store_true",
                      help="Be verbose and chatty on stdout and logfile.")
    parser.add_option("-q", "--quiet", dest="quiet", action="store_true",
                      help="Be quiet, only print warnings and"
                           " errors to stdout and logfile.")
    parser.add_option('', "--version", dest="version", action="store_true",
                      help="Print version and exit.")
    (options, args) = parser.parse_args()
    if options.version:
         print_version_exit("mapwoc-redirector", Redirector.VERSION)
         exit(0) # Not reached.
    if options.quiet and options.verbose:
        print "--verbose takes precedence over --quiet"
    
    setup_logging(options.quiet, options.verbose)
    logger = logging.getLogger('redirector')

    config = ConfigParser.RawConfigParser()
    try:
        config.read(options.config)
        ctrl_port = config.getint('config', 'ctrl_port')
    except ConfigParser.NoSectionError:
        print "Given configuration file %s is invalid." % options.config
        exit(2)

    mapping = {}
    if options.urllistfile is not None:
       mapping = mapping_from_file(options.urllistfile)

    logger.warning("Redirector ignores most of configuration.")
    r = Redirector(ctrl_port=ctrl_port)
    r.start(mapping) 

# Launch main.
if __name__ == "__main__":
    main()

